from collections import defaultdict, deque

import pandas as pd

from OnePy.constants import ActionType
from OnePy.sys_module.metabase_env import OnePyEnvBase


class MatchEngine(OnePyEnvBase):

    def __init__(self, trade_log):
        self.long_log_pure = defaultdict(deque)
        self.long_log_with_trigger = defaultdict(deque)
        self.short_log_pure = defaultdict(deque)
        self.short_log_with_trigger = defaultdict(deque)
        self.finished_log = []
        self.trade_log = trade_log
        self.left_trade_settled = False

    def _append_finished(self, buy_order, sell_order, size):
        log = self.trade_log(buy_order, sell_order, size).generate()
        self.finished_log.append(log)

    def _search_father(self, order, log_with_trigger):
        for i in log_with_trigger:
            if i.mkt_id == order.father_mkt_id:
                self._append_finished(i, order, order.size)
                log_with_trigger.remove(i)

                break

    def _del_in_mkt_dict(self, mkt_id):
        if mkt_id in self.env.orders_pending_mkt_dict:
            del self.env.orders_pending_mkt_dict[mkt_id]

    def _pair_one_by_one(self, order_list, sell_size, order, counteract=False):
        buy_order = order_list.popleft()
        buy_size = buy_order.track_size
        diff = buy_order.track_size = buy_size - sell_size

        if diff > 0:
            self._append_finished(buy_order, order, sell_size)
            order_list.appendleft(buy_order)

            if counteract:  # 修改dict中订单size
                for order in self.env.orders_pending_mkt_dict[buy_order.mkt_id]:
                    order.size = buy_order.track_size

        elif diff == 0:
            self._append_finished(buy_order, order, sell_size)

            if counteract:
                self._del_in_mkt_dict(buy_order.mkt_id)

        else:
            self._append_finished(buy_order, order, buy_size)
            sell_size -= buy_size

            if counteract:
                self._del_in_mkt_dict(buy_order.mkt_id)
            self._pair_one_by_one(order_list, sell_size, order)

    def _pair_order(self, long_or_short, order):  # order should be sell or short cover
        if long_or_short == 'long':
            log_pure = self.long_log_pure[order.ticker]
            log_with_trigger = self.long_log_with_trigger[order.ticker]
        elif long_or_short == 'short':
            log_pure = self.short_log_pure[order.ticker]
            log_with_trigger = self.short_log_with_trigger[order.ticker]

        sell_size = order.size

        if order.father_mkt_id:
            self._search_father(order, log_with_trigger)

        else:

            try:
                self._pair_one_by_one(log_pure, sell_size, order)
            except IndexError:
                self._pair_one_by_one(log_with_trigger, sell_size, order, True)

    def match_order(self, order):
        if order.action_type == ActionType.Buy:
            order.track_size = order.size

            if order.is_pure():
                self.long_log_pure[order.ticker].append(order)
            else:
                self.long_log_with_trigger[order.ticker].append(order)
        elif order.action_type == ActionType.Short_sell:
            order.track_size = order.size

            if order.is_pure():
                self.short_log_pure[order.ticker].append(order)
            else:
                self.short_log_with_trigger[order.ticker].append(order)

        elif order.action_type == ActionType.Sell:
            self._pair_order('long', order)

        elif order.action_type == ActionType.Short_cover:
            self._pair_order('short', order)

    def append_left_trade_to_log(self):

        def settle_left_trade(unfinished_order):
            log = self.trade_log(
                unfinished_order, None, unfinished_order.track_size).settle_left_trade()
            self.finished_log.append(log)

        for ticker in self.env.tickers:

            for order in self.long_log_pure[ticker]:
                settle_left_trade(order)

            for order in self.long_log_with_trigger[ticker]:
                settle_left_trade(order)

            for order in self.short_log_pure[ticker]:
                settle_left_trade(order)

            for order in self.short_log_with_trigger[ticker]:
                settle_left_trade(order)

    def generate_trade_log(self):
        if self.left_trade_settled is False:
            self.append_left_trade_to_log()
            self.left_trade_settled = True

        log_dict = defaultdict(list)

        for log in self.finished_log:
            log_dict['ticker'].append(log.ticker)
            log_dict['entry_date'].append(log.entry_date)
            log_dict['entry_price'].append(log.entry_price)
            log_dict['entry_type'].append(log.entry_type)
            log_dict['size'].append(log.size)
            log_dict['exit_date'].append(log.exit_date)
            log_dict['exit_price'].append(log.exit_price)
            log_dict['exit_type'].append(log.exit_type)
            log_dict['pl_points'].append(log.pl_points)
            log_dict['re_pnl'].append(log.re_pnl)

        return pd.DataFrame(log_dict)



from itertools import count

from dataclasses import dataclass, field

from OnePy.constants import ExecType
from OnePy.sys_module.components.exceptions import (OrderConflictError,
                                                    PctRangeError)
from OnePy.sys_module.metabase_env import OnePyEnvBase


@dataclass
class Signal(OnePyEnvBase):

    counter = count(1)

    size: int
    ticker: str
    datetime: str = None
    takeprofit: float = None
    takeprofit_pct: float = None
    stoploss: float = None
    stoploss_pct: float = None
    trailingstop: float = None
    trailingstop_pct: float = None
    price: float = None
    price_pct: float = None
    execute_price: float = None  # 用来确定是否是必成单,用于挂单
    first_cur_price: float = None

    action_type: str = None
    order_type: str = None
    mkt_id: float = None
    signal_id: int = field(init=False)

    def __post_init__(self):
        self.datetime = self.env.trading_datetime
        self.signal_id = next(self.counter)
        self._check_all_conflict()
        self._save_signals()

    def _save_signals(self):
        self.env.signals_normal_cur.append(self)
        self.env.signals_normal.append(self)

    def _check_all_conflict(self):
        self._check_size()
        self._check_conflict(self.price, self.price_pct, name='price')
        self._check_conflict(
            self.takeprofit, self.takeprofit_pct, name='takeprofit')
        self._check_conflict(self.stoploss, self.stoploss_pct, name='stoploss')
        self._check_conflict(
            self.trailingstop, self.trailingstop_pct, name='trailingstop')

    def _check_size(self):
        if self.size <= 0:
            raise Exception("size should be Positive")

    @staticmethod
    def _check_conflict(obj, obj_pct, name):
        if obj and obj_pct:
            raise OrderConflictError("$ and pct can't be set together")

        if obj_pct:
            if not -1 < obj_pct < 1:
                raise PctRangeError("pct should be -1 < pct < 1")

        if name != 'price':
            if obj:
                if obj <= 0:
                    raise ValueError(f"{name.upper()} should be Positive")

            if obj_pct:
                if obj_pct <= 0:
                    raise ValueError(f"{name.upper()} should be Positive")

    def is_absolute_signal(self):
        return True if self.execute_price else False

    def get(self, name):
        return getattr(self, name)

    def set(self, name, value):
        setattr(self, name, value)


@dataclass
class SignalByTrigger(Signal):
    counter = count(1)

    exec_type: str = None
    trigger_key: str = None

    def _save_signals(self):
        self.env.signals_trigger_cur.append(self)
        self.env.signals_trigger.append(self)



import arrow

from OnePy.event import EVENT, Event
from OnePy.sys_module.components.exceptions import (BacktestFinished,
                                                    BlowUpError)
from OnePy.sys_module.metabase_env import OnePyEnvBase


class MarketMaker(OnePyEnvBase):

    def update_market(self):
        try:
            self._update_bar()
            self._check_todate()
            self._update_recorder()
            self._check_blowup()
            self.env.event_bus.put(Event(EVENT.Market_updated))

        except (StopIteration, BlowUpError):
            self._update_recorder(final=True)  # 最后回测结束用close更新账户信息
            raise BacktestFinished

    def initialize(self):  # 系统初始化
        self._initialize_feeds()
        self._initialize_cleaners()

    def _initialize_feeds(self):  # 初始化行情
        for key, value in self.env.readers.items():
            self.env.feeds.update({key: value.get_bar()})

    def _initialize_cleaners(self):  # 初始化数据给cleaners处理
        for value in self.env.cleaners.values():
            value.initialize_buffer_data()

    def _update_recorder(self, final=False):  # 根据最新行情更新账户信息
        for recorder in self.env.recorders.values():
            recorder.update(final)

    def _update_bar(self):  # 更新行情
        for iter_bar in self.env.feeds.values():
            iter_bar.next()

    def _check_todate(self):  # 检查是否结束回测
        if self.env.todate:
            if arrow.get(self.env.trading_datetime) > arrow.get(self.env.todate):
                raise StopIteration

    def _check_blowup(self):  # 检查是否爆仓
        if self.env.recorder.balance.latest() < 0:
            self.env.logger.critical("The account is BLOW UP!")
            raise BlowUpError




from OnePy.constants import ActionType
from OnePy.sys_module.components.signal_generator import SignalGenerator
from OnePy.sys_module.metabase_env import OnePyEnvBase


class StrategyBase(OnePyEnvBase):

    def __init__(self):
        self._signal_list = []
        self.env.strategies.update({self.__class__.__name__: self})

        self.buy = SignalGenerator(ActionType.Buy).func_1
        self.sell = SignalGenerator(ActionType.Sell).func_2
        self.short_sell = SignalGenerator(ActionType.Short_sell).func_1
        self.short_cover = SignalGenerator(ActionType.Short_cover).func_2

    @property
    def recorder(self):
        """便于在strategy中调用账户信息数据"""

        return self.env.recorder

    def handle_bar(self):
        """在这里写策略的主要逻辑"""
        raise NotImplementedError

    def run(self):
        self.handle_bar()



from OnePy.sys_module.metabase_env import OnePyEnvBase


class BarBase(OnePyEnvBase):

    def __init__(self, reader):
        self._iter_data = reader.load()
        self.current_ohlc = None
        self.next_ohlc = next(self._iter_data)
        self.ticker = reader.ticker
        self._execute_mode = self.env.execute_on_close_or_next_open

    def next(self): # 更新行情
        self.current_ohlc, self.next_ohlc = self.next_ohlc, next(
            self._iter_data)

    @property
    def cur_price(self):
        return self.close

    @property
    def execute_price(self):
        if self._execute_mode == 'open':
            return self.next_ohlc['open']

        return self.close

    @property
    def date(self):
        return self.current_ohlc['date']

    @property
    def open(self):
        return self.current_ohlc['open']

    @property
    def high(self):
        return self.current_ohlc['high']

    @property
    def low(self):
        return self.current_ohlc['low']

    @property
    def close(self):
        return self.current_ohlc['close']

    @property
    def volume(self):
        return self.current_ohlc['volume']



from OnePy.builtin_module.trade_log.match_engine import MatchEngine
from OnePy.sys_module.metabase_env import OnePyEnvBase


class RecorderBase(OnePyEnvBase):

    def __init__(self, trade_log):
        self.env.recorders.update({self.__class__.__name__: self})
        self.env.recorder = self
        self.match_engine = MatchEngine(trade_log)

        self.initial_cash = 100000
        self.per_comm = 1
        self.per_comm_pct = None
        self.margin_rate = 0.1

        self.position = None
        self.avg_price = None
        self.holding_pnl = None
        self.realized_pnl = None
        self.commission = None
        self.market_value = None
        self.margin = None

        self.cash = None
        self.frozen_cash = None
        self.balance = None

    def _record_order(self):
        """记录成交的账单信息，更新账户信息"""
        raise NotImplementedError

    def run(self):
        raise NotImplementedError

    def update(self):
        """根据最新价格更新账户信息"""
        raise NotImplementedError



from itertools import count

from OnePy.constants import ActionType
from OnePy.sys_module.metabase_env import OnePyEnvBase
from OnePy.sys_module.models.orders.general_order import (LimitBuyOrder,
                                                          LimitCoverShortOrder,
                                                          LimitSellOrder,
                                                          LimitShortSellOrder,
                                                          MarketOrder,
                                                          StopBuyOrder,
                                                          StopCoverShortOrder,
                                                          StopSellOrder,
                                                          StopShortSellOrder,
                                                          TrailingStopCoverShortOrder,
                                                          TrailingStopSellOrder)


class OrderGenerator(OnePyEnvBase):

    counter = count(1)

    def __init__(self):
        self.signal = None
        self.mkt_id = None

        self.market_order = None
        self.orders_pending_mkt = None
        self.orders_pending = None

    @property
    def cur_price(self):
        return self.env.feeds[self.signal.ticker].cur_price

    def is_buy(self):
        return True if self.signal.action_type == ActionType.Buy else False

    def is_sell(self):
        return True if self.signal.action_type == ActionType.Sell else False

    def is_shortsell(self):
        return True if self.signal.action_type == ActionType.Short_sell else False

    def is_shortcover(self):
        return True if self.signal.action_type == ActionType.Short_cover else False

    def is_exitall(self):
        return True if self.signal.action_type == ActionType.Exit_all else False

    def is_cancelall(self):
        return True if self.signal.action_type == ActionType.Cancel_all else False

    def is_absolute_mkt(self):
        return True if self.signal.is_absolute_signal() else False

    def is_normal_mkt(self):
        return False if self.signal.price or self.signal.price_pct else True

    def is_marketorder(self):
        if self.is_absolute_mkt() or self.is_normal_mkt():
            return True

        return False

    def _set_market_order(self):
        self.market_order = MarketOrder(self.signal, self.mkt_id)

    def _clarify_price_pct(self):
        if self.signal.price_pct:
            self.signal.price = (self.signal.price_pct+1)*self.cur_price

    def _child_of_mkt(self, order_class, key):
        if self.signal.get(key):
            self.orders_pending_mkt.append(
                order_class(self.signal, self.mkt_id, key))
        elif self.signal.get(f'{key}_pct'):
            self.orders_pending_mkt.append(
                order_class(self.signal, self.mkt_id, f'{key}_pct'))

    def _pending_order_only(self, order_class):
        self.orders_pending.append(order_class(
            self.signal, None, 'price'))

    def _generate_child_order_of_mkt(self):

        if self.is_buy():
            self._child_of_mkt(StopSellOrder, 'stoploss')
            self._child_of_mkt(LimitSellOrder, 'takeprofit')
            self._child_of_mkt(TrailingStopSellOrder, 'trailingstop')

        elif self.is_shortsell():
            self._child_of_mkt(StopCoverShortOrder, 'stoploss')
            self._child_of_mkt(LimitCoverShortOrder, 'takeprofit')
            self._child_of_mkt(
                TrailingStopCoverShortOrder, 'trailingstop')

    def _generate_pending_order_only(self):
        self._clarify_price_pct()

        if self.signal.price > self.cur_price:
            if self.is_buy():
                self._pending_order_only(StopBuyOrder)
            elif self.is_shortcover():
                self._pending_order_only(StopCoverShortOrder)
            elif self.is_sell():
                self._pending_order_only(LimitSellOrder)
            elif self.is_shortsell():
                self._pending_order_only(LimitShortSellOrder)
        elif self.signal.price < self.cur_price:
            if self.is_buy():
                self._pending_order_only(LimitBuyOrder)
            elif self.is_shortcover():
                self._pending_order_only(LimitCoverShortOrder)
            elif self.is_sell():
                self._pending_order_only(StopSellOrder)
            elif self.is_shortsell():
                self._pending_order_only(StopShortSellOrder)
        else:
            self.signal.execute_price = self.cur_price
            self._generate_order()

    def _initialize(self, signal):
        self.signal = signal
        self.market_order = None
        self.orders_pending_mkt = []
        self.orders_pending = []

    def _generate_order(self):

        if self.is_marketorder():
            self.mkt_id = next(self.counter)
            self._set_market_order()
            self._generate_child_order_of_mkt()

        else:
            self._generate_pending_order_only()

    def _submit_order_to_env(self):
        if self.market_order:
            self.env.orders_mkt_original.append(self.market_order)

            if self.is_absolute_mkt():
                self.env.orders_mkt_absolute.append(self.market_order)
            elif self.is_normal_mkt():
                self.env.orders_mkt_normal.append(self.market_order)

            if self.orders_pending_mkt != []:
                self.env.orders_pending_mkt_dict.update(
                    {self.mkt_id: self.orders_pending_mkt})
        else:
            self.env.orders_pending += self.orders_pending

    def _process_every_signals_in(self, signals_list):
        for signal in signals_list:
            self._initialize(signal)
            self._generate_order()
            self._submit_order_to_env()

    def _clear_current_signals_memory(self):
        self.env.signals_normal_cur = []
        self.env.signals_trigger_cur = []

    def run(self):
        self._process_every_signals_in(self.env.signals_normal_cur)
        self._process_every_signals_in(self.env.signals_trigger_cur)
        self._clear_current_signals_memory()



from OnePy.event import EventBus


class Environment(object):

    """作为全局共享变量为各模块提供支持"""

    fromdate = None
    todate = None
    event_bus = EventBus()

    mod_dict = {}
    readers = {}
    feeds = {}
    cleaners = {}
    strategies = {}
    brokers = {}
    risk_managers = {}
    recorders = {}
    recorder = None
    logger = None

    signals_normal: list = []  # 保存最原始的所有信号
    signals_trigger: list = []  # 保存最原始的所有挂单信号
    signals_normal_cur: list = []  # 动态地临时信号，会不断刷新
    signals_trigger_cur: list = []  # 动态地临时信号，会不断刷新

    orders_mkt_original: list = []  # 保存最原始的所有订单信号
    orders_mkt_normal: list = []  # 动态地保存当前订单, 会不断刷新
    orders_mkt_absolute: list = []  # 动态地保存触发的挂单并成交信息，会不断刷新
    orders_mkt_submitted: list = []  # 动态地保存成交单，会不断刷新

    orders_pending: list = []   # 动态地保存挂单
    orders_pending_mkt_dict: dict = {}  # 动态地保存跟随市价单的挂单

    event_loop = None
    execute_on_close_or_next_open = 'open'

    def refresh(self):
        """刷新environment防止缓存累积"""
        self.signals_normal: list = []
        self.signals_trigger: list = []
        self.signals_normal_cur: list = []
        self.signals_trigger_cur: list = []
        self.orders_mkt_original: list = []
        self.orders_mkt_normal: list = []
        self.orders_mkt_absolute: list = []
        self.orders_mkt_submitted: list = []
        self.orders_pending: list = []
        self.orders_pending_mkt_dict: dict = {}

    @property
    def trading_datetime(self):
        return self.feeds[self.tickers[0]].date

    @property
    def tickers(self):
        return list(self.readers.keys())



import json

import tushare as ts

from OnePy.builtin_module.mongodb_saver.mongodb_config import MongodbConfig


class Tushare_to_MongoDB(MongodbConfig):
    host = "localhost"
    port = 27017
    dtformat = "%Y-%m-%d %H:%M"
    tmformat = "%H:%M:%S"
    date = "date"
    time = None
    open = "open"
    high = "high"
    low = "low"
    close = "close"
    volume = "volume"
    openinterest = None

    def __init__(self, database, collection, host=None, port=None):
        super(Tushare_to_MongoDB, self).__init__(
            database, collection, host, port)

    def __load_csv(self, code, start, end, ktype, autype, index):
        d = dict(code=code, ktype=ktype)

        if start:
            d["start"] = start

        if end:
            d["end"] = end

        if ktype is "D":
            self.dtformat = "%Y-%m-%d"
        df = ts.get_k_data(**d)
        df.reset_index(drop=True, inplace=True)
        j = df.to_json()
        data = json.loads(j)

        return data

    def data_to_db(self, code, start=None, end=None,
                   ktype="D", autype="qfq", index=False):
        data = self.__load_csv(code, start, end, ktype, autype, index)
        self._combine_and_insert(data)
        self._drop_duplicates()




import pymongo

from OnePy.sys_module.base_reader import DataReaderBase
from OnePy.sys_module.models.bar_backtest import BarBacktest


class MongodbReader(DataReaderBase):
    host = 'localhost'
    port = 27017

    def __init__(self, database, collection, ticker, fromdate=None, todate=None, host=None, port=None):
        super().__init__(ticker, fromdate, todate)
        self.host = host if host else self.host
        self.port = port if port else self.port

        self.database = database
        self.collection = collection

    def set_collection(self):
        client = pymongo.MongoClient(host=self.host, port=self.port)
        db = client[self.database]
        Collection = db[self.collection]

        return Collection

    def load(self, fromdate=None, todate=None):
        coll = self.set_collection()

        if fromdate is None:
            fromdate = self.fromdate

        if todate is None:
            todate = self.todate

        if fromdate and todate:
            return coll.find({'date': {'$gte': fromdate, '$lt': todate}})
        elif fromdate:
            return coll.find({'date': {'$gte': fromdate}})
        elif todate:
            return coll.find({'date': {'$lt': todate}})
        else:
            return coll.find()

    def get_bar(self):
        return BarBacktest(self)



from collections import defaultdict, deque
from functools import partial
from itertools import count

import arrow

from OnePy.sys_module.metabase_env import OnePyEnvBase


class CleanerBase(OnePyEnvBase):
    cleaner_counter = count(1)

    def __init__(self, rolling_window, buffer_day):
        self.name = f'{self.__class__.__name__}_{next(self.cleaner_counter)}'
        self.env.cleaners.update({self.name: self})
        self.rolling_window = rolling_window
        self.buffer_day = buffer_day
        self.retry = count(1)
        self.data = None

    @property
    def startdate(self):
        date = arrow.get(self.env.fromdate).shift(days=-self.buffer_day)

        return date.format('YYYY-MM-DD HH:mm:ss')

    def _check_length(self, ticker):
        if len(self.data[ticker]) < self.data[ticker].maxlen:

            self.buffer_day += 2
            self.initialize_buffer_data()

            self.env.logger.warning(
                f'{self.name} buffer_day is too short! Retry No.{next(self.retry)} times')

    def initialize_buffer_data(self):
        self.data = defaultdict(partial(deque, maxlen=self.rolling_window))

        for key, value in self.env.readers.items():
            buffer_data = value.load(
                fromdate=self.startdate, todate=self.env.fromdate)

            self.data[key].extend((i['close'] for i in buffer_data))
            self._check_length(key)

    def _append_data_to_buffer(self):
        for ticker, ohlc in self.env.feeds.items():
            self.data[ticker].append(ohlc.close)

    def run(self):
        self._append_data_to_buffer()

    def calculate(self, ticker):
        raise NotImplementedError



from OnePy.constants import ActionType
from OnePy.sys_module.metabase_env import OnePyEnvBase
from OnePy.sys_module.models.signals import Signal, SignalByTrigger


class SignalGenerator(OnePyEnvBase):
    def __init__(self, action_type):
        self.action_type = action_type

    def func_1(self, size, ticker,
               takeprofit=None, takeprofit_pct=None,
               stoploss=None, stoploss_pct=None,
               trailingstop=None, trailingstop_pct=None,
               price=None, price_pct=None):  # For Buy, ShortSell

        return Signal(
            action_type=self.action_type,
            size=size,
            ticker=ticker,
            takeprofit=takeprofit,
            takeprofit_pct=takeprofit_pct,
            stoploss=stoploss,
            stoploss_pct=stoploss_pct,
            trailingstop=trailingstop,
            trailingstop_pct=trailingstop_pct,
            price=price,
            price_pct=price_pct,
        )

    def func_2(self, size, ticker, price=None, price_pct=None):  # For Sell, ShortCover

        return Signal(
            action_type=self.action_type,
            size=size,
            ticker=ticker,
            price=price,
            price_pct=price_pct,
        )


class TriggeredSignalGenerator:
    """为触发的挂单生成挂单触发信号"""
    @classmethod
    def _generate_bare_signal(cls, order):
        return SignalByTrigger(
            action_type=order.action_type,
            size=order.size,
            ticker=order.ticker,
            execute_price=order.target_price,
            first_cur_price=order.first_cur_price,
            mkt_id=order.mkt_id,
            order_type=order.order_type)

    @classmethod
    def _generate_full_signal(cls, order):
        return SignalByTrigger(action_type=order.action_type,
                               size=order.size,
                               ticker=order.ticker,
                               execute_price=order.target_price,
                               price=None, price_pct=None,
                               takeprofit=order.signal.takeprofit,
                               takeprofit_pct=order.signal.takeprofit_pct,
                               stoploss=order.signal.stoploss,
                               stoploss_pct=order.signal.stoploss_pct,
                               trailingstop=order.signal.trailingstop,
                               trailingstop_pct=order.signal.trailingstop_pct,
                               order_type=order.order_type)

    @classmethod
    def generate_triggered_signal(cls, order):
        if order.is_triggered:
            if order.is_with_mkt():
                return cls._generate_bare_signal(order)

            return cls._generate_full_signal(order)




import pandas as pd
from plotly import graph_objs as go
from plotly import offline as py

from OnePy.sys_module.metabase_env import OnePyEnvBase


class PlotBase(OnePyEnvBase):

    def __init__(self):
        super().__init__()
        self.balance_df = self.env.recorder.balance.dataframe()
        self.cash_df = self.env.recorder.cash.dataframe()
        self.positions_df = self.env.recorder.position.dataframe()
        self.holding_pnl_df = self.env.recorder.holding_pnl.dataframe()
        self.commission_df = self.env.recorder.commission.dataframe()
        self.margin_df = self.env.recorder.margin.dataframe()
        self.data = []
        self.updatemenus = []

    @property
    def realized_pnl_df(self):
        trade_log = self.env.recorder.match_engine.generate_trade_log()
        trade_log.dropna(inplace=True)
        df = trade_log[['exit_date', 're_pnl']].copy()
        df.rename(columns=dict(exit_date='date'), inplace=True)
        df.set_index('date', drop=True, inplace=True)
        df.index = pd.to_datetime(df.index)

        return df

    def ohlc_df(self, ticker):
        ohlc = self.env.readers[ticker].load(
            self.env.fromdate, self.env.todate)
        dataframe = pd.DataFrame((i for i in ohlc))
        dataframe.set_index('date', inplace=True)
        dataframe.index = pd.to_datetime(dataframe.index)

        return dataframe


class Plotly(PlotBase):

    def plot(self, instrument=None, engine='plotly', notebook=False):

        if engine == 'plotly':
            if isinstance(instrument, str):
                df = self.ohlc_df(instrument)
                p_symbol = go.Scatter(x=df.index, y=df.close,
                                      xaxis='x3', yaxis='y3', name=instrument)
                p_volume = go.Bar(x=df.index, y=df['volume'],
                                  xaxis='x3', yaxis='y5', opacity=0.5, name='volume')
                self.data.append(p_symbol)
                self.data.append(p_volume)

            if isinstance(instrument, list):
                for i in instrument:
                    df = self.ohlc_df(instrument)
                    p_symbol = go.Scatter(x=df.index, y=df.close,
                                          xaxis='x3', yaxis='y3', name=i)
                    p_volume = go.Bar(x=df.index, y=df['volume'],
                                      xaxis='x3', yaxis='y5',
                                      opacity=0.5, name=i + 'volume')
                    self.data.append(p_symbol)
                    self.data.append(p_volume)

        for i in self.positions_df:
            self.positions_df.sort_index(inplace=True)
            p_position = go.Scatter(x=self.positions_df.index,
                                    y=self.positions_df[i],
                                    xaxis='x2', yaxis='y2', name=i)
            self.data.append(p_position)

        for i in self.realized_pnl_df:
            self.realized_pnl_df.sort_index(inplace=True)
            p_re_profit = go.Scatter(x=self.realized_pnl_df.index,
                                     y=self.realized_pnl_df[i],
                                     xaxis='x4', yaxis='y4',
                                     name=i)

            self.data.append(p_re_profit)

        for i in self.holding_pnl_df:
            self.holding_pnl_df.sort_index(inplace=True)

            p_unre_profit = go.Scatter(x=self.holding_pnl_df.index,
                                       y=self.holding_pnl_df[i],
                                       xaxis='x4', yaxis='y4',
                                       name=i)

            self.data.append(p_unre_profit)

        for i in self.commission_df:
            self.commission_df.sort_index(inplace=True)
            p_commission = go.Scatter(x=self.commission_df.index,
                                      y=self.commission_df[i],
                                      xaxis='x4', yaxis='y4',
                                      name=i, visible='legendonly')

            self.data.append(p_commission)
        p_total = go.Scatter(x=self.balance_df.index,
                             y=self.balance_df.balance,
                             xaxis='x6', yaxis='y6', name='balance')

        p_cash = go.Scatter(x=self.cash_df.index,
                            y=self.cash_df.cash,
                            xaxis='x6', yaxis='y6', name='cash')

        self.data.append(p_total)
        self.data.append(p_cash)

        layout = go.Layout(
            xaxis2=dict(
                domain=[0, 1],
                anchor='y2',
                scaleanchor='x2'
            ),
            xaxis3=dict(
                domain=[0, 1],
                anchor='y3',
                scaleanchor='x2'
            ),
            xaxis4=dict(
                domain=[0, 1],
                anchor='y4',
                scaleanchor='x2'
            ),
            xaxis6=dict(
                domain=[0, 1],
                anchor='y6',
                scaleanchor='x2'
            ),
            yaxis2=dict(
                domain=[0, 0.15],
                scaleanchor='x2'
            ),
            yaxis3=dict(
                domain=[0.15, 0.65],
                scaleanchor='x2'
            ),
            yaxis4=dict(
                domain=[0.65, 0.85],
                scaleanchor='x2'
            ),
            yaxis5=dict(
                domain=[0.15, 0.65],
                side='right',
                range=[0, 10000000],
                overlaying='y3',
                tickvals=[0, 1000000, 2000000, 2500000],
                showgrid=False,
                scaleanchor='x2'
            ),
            yaxis6=dict(
                domain=[0.85, 1],
                scaleanchor='x2'
            )
        )
        fig = go.Figure(data=self.data, layout=layout)

        if notebook:
            import plotly
            plotly.offline.init_notebook_mode()
            py.iplot(fig, filename='OnePy_plot.html', validate=False)
        else:
            py.plot(fig, filename='OnePy_plot.html', validate=False)



from dataclasses import dataclass, field

from OnePy.constants import ActionType
from OnePy.sys_module.metabase_env import OnePyEnvBase


@dataclass
class TradeLogBase(OnePyEnvBase):

    buy: float = None
    sell: float = None
    size: float = None

    ticker: str = field(init=False)

    entry_date: str = field(init=False)
    exit_date: str = field(init=False)

    entry_price: float = field(init=False)
    exit_price: float = field(init=False)
    pl_points: float = field(init=False)
    re_pnl: float = field(init=False)

    # commission: float = field(init=False)
    # cumulative_total: float = field(init=False)

    def generate(self):
        raise NotImplementedError

    def _earn_short(self):
        return -1 if self.buy.action_type == ActionType.Short_sell else 1

    def _get_order_type(self, order):
        if order.signal.order_type:
            return order.signal.order_type.value
        else:
            return order.order_type.value

    def settle_left_trade(self):
        raise NotImplementedError



from OnePy.constants import ActionType, OrderType
from OnePy.sys_module.models.orders.base_order import (OrderBase,
                                                       PendingOrderBase,
                                                       TrailingOrderBase)


class MarketOrder(OrderBase):

    order_type = OrderType.Market

    @property
    def execute_price(self):
        return self.first_cur_price

    def is_pure(self):
        return False if self.mkt_id in self.env.orders_pending_mkt_dict else True

    @property
    def father_mkt_id(self):
        return self.signal.mkt_id

    @property
    def status(self):
        return self._status

    @status.setter
    def status(self, value):
        self._status = value
        self.env.logger.info(
            f'{self.signal.datetime}, '
            f'{self.signal.ticker}, '
            f'{self.order_type.value} '
            f'{self.action_type.value} '
            f'@ {self.execute_price:.5f}, '
            f'{self.status.value}, '
            f'size: {self.size}')


class LimitBuyOrder(PendingOrderBase):
    """
    如果是挂单，只需要判断是不是触发，
    如果触发，就返回执行市价单，而且是必成的那种,而且带着其他挂单
    如果是跟随mkt的挂单，则需要判断是否触发，
        如果触发，就发送一个裸的必成市价单，即要清空signal，剩裸price

    而且如果触发了，要从list中删除
    """
    _action_type = ActionType.Buy
    order_type = OrderType.Limit

    @property
    def target_below(self):
        return True


class LimitSellOrder(PendingOrderBase):
    _action_type = ActionType.Sell
    order_type = OrderType.Limit

    @property
    def target_below(self):
        return False


class StopBuyOrder(LimitSellOrder):
    _action_type = ActionType.Buy
    order_type = OrderType.Stop


class StopSellOrder(LimitBuyOrder):
    _action_type = ActionType.Sell
    order_type = OrderType.Stop


class LimitShortSellOrder(LimitSellOrder):
    _action_type = ActionType.Short_sell
    order_type = OrderType.Limit


class StopShortSellOrder(StopSellOrder):
    _action_type = ActionType.Short_sell
    order_type = OrderType.Stop


class LimitCoverShortOrder(LimitBuyOrder):
    _action_type = ActionType.Short_cover
    order_type = OrderType.Limit


class StopCoverShortOrder(StopBuyOrder):
    _action_type = ActionType.Short_cover
    order_type = OrderType.Stop


class TrailingStopSellOrder(TrailingOrderBase):
    _action_type = ActionType.Sell
    order_type = OrderType.Trailing_stop

    @property
    def target_below(self):
        return True


class TrailingStopCoverShortOrder(TrailingOrderBase):
    _action_type = ActionType.Short_cover
    order_type = OrderType.Trailing_stop

    @property
    def target_below(self):
        return False



import json
from datetime import datetime

import funcy as fy
import pandas as pd
import pymongo


class MongodbConfig(object):
    host = 'localhost'
    port = 27017
    dtformat = '%Y-%m-%d %H:%M:%S'
    tmformat = '%H:%M:%S'
    date = 'Date'
    time = 'Timestamp'
    open = 'Open'
    high = 'High'
    low = 'Low'
    close = 'Close'
    volume = 'Volume'
    openinterest = None

    def __init__(self, database, collection, host=None, port=None):
        self.host = host if host else self.host
        self.port = port if port else self.port
        self.database = database
        self.collection = collection

    def __set_dtformat(self, bar):
        """识别日期"""
        date = bar[self.date.lower()]
        dt = "%Y-%m-%d %H:%M:%S"

        if '%H' in self.dtformat:
            return datetime.strptime(str(date), self.dtformat).strftime(dt)
        elif self.time:
            date = datetime.strptime(
                str(date), self.dtformat).strftime('%Y-%m-%d')

            return date + ' ' + bar[self.time.lower()]
        else:
            return datetime.strptime(str(date), self.dtformat).strftime('%Y-%m-%d')

    def _set_collection(self):
        """设置数据库"""
        client = pymongo.MongoClient(host=self.host, port=self.port)
        db = client[self.database]
        Collection = db[self.collection]

        return Collection

    def __load_csv(self, path):
        """读取CSV"""
        df = pd.read_csv(path)
        j = df.to_json()
        data = json.loads(j)

        return data

    def _combine_and_insert(self, data):
        """整合并插入数据"""
        # 构造 index 列表
        name_list = [self.date, self.time, self.open, self.high,
                     self.low, self.close, self.volume, self.openinterest]
        # 删除 None

        for i in range(len(name_list)):
            if None in name_list:
                name_list.remove(None)

        def process_data(n):
            # 返回单个数据的字典，key为index，若无index则返回 None
            single_data = {index.lower(): data[index].get(str(n))
                           for index in name_list}

            return single_data

        lenth = len(data[self.date])  # 总长度
        coll = self._set_collection()

        # 插入数据

        for i in range(lenth):
            bar = process_data(i)
            bar[self.date.lower()] = self.__set_dtformat(bar)
            coll.insert_one(bar)
            print('Inserting ' + str(i) + ', Total: ' + str(lenth))

    def __get_dups_id(self, data):
        """获得重复数据的id"""
        data['dups'].pop(0)

        return data['dups']

    def _drop_duplicates(self):
        """删除重复数据"""
        coll = self._set_collection()
        c = coll.aggregate([{"$group":
                             {"_id": {'date': '$date'},
                              "count": {'$sum': 1},
                              "dups": {'$addToSet': '$_id'}}},
                            {'$match': {'count': {"$gt": 1}}}
                            ]
                           )
        data = [i for i in c]
        duplicates = fy.walk(self.__get_dups_id, data)
        dups_id_list = fy.cat(duplicates)

        for i in dups_id_list:
            coll.delete_one({'_id': i})
        # print("OK, duplicates droped! Done!")

    def data_to_db(self, path):
        """数据导入数据库"""
        data = self.__load_csv(path)
        self._combine_and_insert(data)
        self._drop_duplicates()



from OnePy.constants import ActionType
from OnePy.sys_module.models.base_series import SeriesBase


class PositionSeries(SeriesBase):
    name = 'position'

    def append(self, order, last_position,  long_or_short='long'):
        new_value = last_position + order.size*self.direction(order)
        self._append_value(order.ticker, new_value, long_or_short)


class AvgPriceSeries(SeriesBase):
    name = 'avg_price'

    def append(self, order, last_position, last_avg_price, new_position,
               long_or_short='long'):

        if new_position == 0:
            new_value = 0

        elif new_position > last_position:
            new_value = (last_position * last_avg_price +
                         order.size*order.execute_price)/new_position
        else:
            new_value = last_avg_price  # TODO 错误

        self._append_value(order.ticker, new_value, long_or_short)


class RealizedPnlSeries(SeriesBase):
    """
    1. Buy and Sell对冲
    2. short sell 和short cover
    要添加负号
    """
    name = 'realized_pnl'

    def append(self, order, last_avg_price, new_avg_price, long_or_short='long'):
        new_value = None

        if order.action_type == ActionType.Sell:
            new_value = self.latest(
                order.ticker, long_or_short)+(order.execute_price - last_avg_price)*order.size
        elif order.action_type == ActionType.Short_cover:
            new_value = self.latest(
                order.ticker, long_or_short)-(order.execute_price - last_avg_price)*order.size

        if new_value:
            self._append_value(order.ticker,  new_value, long_or_short)


class CommissionSeries(SeriesBase):
    name = 'commission'

    def append(self, order, last_commission, per_comm, per_comm_pct,
               long_or_short='long'):

        if per_comm_pct:
            new_value = last_commission + per_comm*order.size*order.execute_price
        else:
            new_value = last_commission + per_comm
        self._append_value(order.ticker, new_value, long_or_short)


class HoldingPnlSeries(SeriesBase):
    name = 'holding_pnl'

    def append(self, ticker, cur_price, new_avg_price, new_position,
               long_or_short='long'):

        if new_position == 0:
            new_value = 0
        else:
            new_value = (cur_price - new_avg_price)*new_position * \
                self.earn_short(long_or_short)
        self._append_value(ticker, new_value, long_or_short)

    def update_barly(self, order_executed):
        for ticker in self.env.feeds:
            cur_price = self.get_barly_cur_price(ticker, order_executed)

            for long_or_short in ['long', 'short']:
                new_position = self.env.recorder.position.latest(
                    ticker, long_or_short)
                new_avg_price = self.env.recorder.avg_price.latest(
                    ticker, long_or_short)

                self.append(ticker, cur_price, new_avg_price, new_position,
                            long_or_short)


class MarketValueSeries(SeriesBase):
    name = 'market_value'

    def append(self, ticker,  cur_price, new_position,
               long_or_short='long'):
        new_value = new_position*cur_price
        self._append_value(ticker,  new_value, long_or_short)

    def update_barly(self, order_executed):
        for ticker in self.env.feeds:
            cur_price = self.get_barly_cur_price(ticker, order_executed)

            for long_or_short in ['long', 'short']:
                new_position = self.env.recorder.position.latest(
                    ticker, long_or_short)
                self.append(ticker,  cur_price,
                            new_position, long_or_short)


class MarginSeries(SeriesBase):
    name = 'margin'

    def append(self, ticker,  cur_price, new_position, margin_rate,
               long_or_short='long'):

        if long_or_short == 'short':
            new_value = new_position*cur_price*margin_rate
            self._append_value(
                ticker,  new_value, long_or_short)

    def update_barly(self, order_executed):

        for ticker in self.env.feeds:
            cur_price = self.get_barly_cur_price(ticker, order_executed)

            for long_or_short in ['long', 'short']:
                new_position = self.env.recorder.position.latest(
                    ticker, long_or_short)

                margin_rate = self.env.recorder.margin_rate
                self.append(ticker,  cur_price, new_position, margin_rate,
                            long_or_short)



from dataclasses import dataclass, field

from OnePy.builtin_module.trade_log.base_log import TradeLogBase
from OnePy.constants import ActionType


@dataclass
class StockTradeLog(TradeLogBase):
    buy: float = None
    sell: float = None
    size: float = None
    ticker: str = field(init=False)
    entry_date: str = field(init=False)
    exit_date: str = field(init=False)
    entry_price: float = field(init=False)
    exit_price: float = field(init=False)
    pl_points: float = field(init=False)
    re_pnl: float = field(init=False)

    def generate(self):
        sell_order_type = self._get_order_type(self.sell)
        buy_order_type = self._get_order_type(self.buy)

        self.ticker = self.buy.ticker
        self.entry_date = self.buy.trading_date
        self.entry_price = self.buy.first_cur_price
        self.entry_type = f'{buy_order_type} {self.buy.action_type.value}'
        self.exit_date = self.sell.trading_date
        self.exit_price = self.sell.first_cur_price
        self.exit_type = f'{sell_order_type} {self.sell.action_type.value}'
        self.pl_points = (self.sell.first_cur_price -
                          self.buy.first_cur_price)*self._earn_short()
        self.re_pnl = self.pl_points*self.size

        return self

    def settle_left_trade(self):
        ticker = self.buy.ticker
        cur_price = self.env.feeds[ticker].cur_price
        buy_order_type = self._get_order_type(self.buy)

        self.ticker = ticker
        self.entry_date = self.buy.trading_date
        self.entry_price = self.buy.first_cur_price
        self.entry_type = f'{buy_order_type} {self.buy.action_type.value}'
        self.exit_date = None
        self.exit_price = None
        self.exit_type = None
        self.pl_points = (
            cur_price - self.buy.execute_price)*self._earn_short()
        self.re_pnl = self.pl_points*self.size

        return self



from OnePy.builtin_module import data_readers
from OnePy.main import OnePiece
from OnePy.sys_module.base_broker import StockBroker
from OnePy.sys_module.base_cleaner import CleanerBase
from OnePy.sys_module.base_reader import DataReaderBase
from OnePy.sys_module.base_recorder import RecorderBase
from OnePy.sys_module.base_riskmanager import RiskManagerBase
from OnePy.sys_module.base_strategy import StrategyBase



import OnePy as op
from OnePy.sys_module.base_cleaner import CleanerBase


class SMA(CleanerBase):

    def calculate(self, ticker):
        close = self.data[ticker]

        return sum(close)/len(close)  # TODO:尝试用numpy看能否提高性能


if __name__ == "__main__":
    db = op.data_readers.MongodbReader(
        'tushare', '000001', '000001', fromdate='2017-02-12')

    a = SMA(10, 21)
    a.initialize_buffer_data()
    a.calculate('000001')
    print(a.data)



from OnePy.builtin_module.mongodb_saver.mongodb_config import MongodbConfig


class Forex_csv_to_MongoDB(MongodbConfig):
    host = 'localhost'
    port = 27017
    dtformat = '%Y%m%d'
    tmformat = '%H:%M:%S'
    date = 'Date'
    time = 'Timestamp'
    open = 'Open'
    high = 'High'
    low = 'Low'
    close = 'Close'
    volume = 'Volume'
    openinterest = None

    def __init__(self, database, collection, host=None, port=None):
        super(Forex_csv_to_MongoDB, self).__init__(
            database, collection, host, port)


class Tushare_csv_to_MongoDB(MongodbConfig):
    host = 'localhost'
    port = 27017
    dtformat = '%Y-%m-%d'
    tmformat = '%H:%M:%S'
    date = 'date'
    time = None
    open = 'open'
    high = 'high'
    low = 'low'
    close = 'close'
    volume = 'volume'
    openinterest = None

    def __init__(self, database, collection, host=None, port=None):
        super(Tushare_csv_to_MongoDB, self).__init__(
            database, collection, host, port)



import csv

import arrow

from OnePy.sys_module.base_reader import DataReaderBase
from OnePy.sys_module.models.bar_backtest import BarBacktest


class CSVReader(DataReaderBase):

    def __init__(self,  data_path, ticker, fromdate=None, todate=None):
        super().__init__(ticker, fromdate, todate)
        self.data_path = data_path

    def get_bar(self):
        return BarBacktest(self)

    def _load_raw_data(self):
        return csv.DictReader(open(self.data_path))

    def load(self, fromdate=None, todate=None):
        if fromdate is None:
            fromdate = self.fromdate

        if todate is None:
            todate = self.todate

        generator = self._load_raw_data()
        final_data = []

        for ohlc in generator:
            if todate:
                if arrow.get(ohlc['date']) >= arrow.get(todate):
                    break

            if arrow.get(ohlc['date']) >= arrow.get(fromdate):
                ohlc['open'] = float(ohlc['open'])
                ohlc['high'] = float(ohlc['high'])
                ohlc['low'] = float(ohlc['low'])
                ohlc['close'] = float(ohlc['close'])
                ohlc['volume'] = float(ohlc['volume'])

                final_data.append(ohlc)
        final_generator = (i for i in final_data)
        del final_data

        return final_generator



from OnePy.sys_module.metabase_env import OnePyEnvBase


class RiskManagerBase(OnePyEnvBase):

    def __init__(self):
        self.env.risk_managers.update({self.__class__.__name__: self})

    def run(self):
        pass



from OnePy.sys_module.metabase_env import OnePyEnvBase


class CashChecker(OnePyEnvBase):

    @classmethod
    def stock_checker(self, order):  # TODO：还要加上手续费,和保证金
        return order.size * order.execute_price



from OnePy.builtin_module.plotters.by_plotly import PlotBase
from OnePy.builtin_module.trade_log.match_engine import MatchEngine
from OnePy.constants import EVENT
from OnePy.environment import Environment

# 控制事件发生顺序
EVENT_LOOP = [dict(if_event=EVENT.Market_updated,
                   then_event=EVENT.Data_cleaned,
                   module_dict=Environment.cleaners),

              dict(if_event=EVENT.Data_cleaned,
                   then_event=EVENT.Signal_generated,
                   module_dict=Environment.strategies),

              dict(if_event=EVENT.Signal_generated,
                   then_event=EVENT.Submit_order,
                   module_dict=Environment.risk_managers),

              dict(if_event=EVENT.Submit_order,
                   then_event=EVENT.Record_result,
                   module_dict=Environment.brokers),

              dict(if_event=EVENT.Record_result,
                   then_event=None,
                   module_dict=Environment.recorders)]




from OnePy.sys_module.metabase_env import OnePyEnvBase


class DataReaderBase(OnePyEnvBase):

    """负责读取数据"""

    def __init__(self, ticker, fromdate=None, todate=None):
        self.ticker = ticker
        self.fromdate = fromdate
        self.todate = todate

        self.env.readers[self.ticker] = self
        self.env.fromdate = fromdate
        self.env.todate = todate

        self.margin_per_lot = None
        self.dollar_per_pips = None

    def load(self, fromdate=None, todate=None):
        """需要返回已过滤好的从fromdate开始的数据,cleander需要用"""
        raise NotImplementedError

    def get_bar(self):
        raise NotImplementedError



import pandas as pd

from OnePy.builtin_module.plotters.by_plotly import Plotly
from OnePy.sys_module.metabase_env import OnePyEnvBase
from OnePy.utils.awesome_func import dict_to_table
from OnePy.utils.easy_func import check_setting
from OnePy.utils.statistics import create_drawdowns, create_sharpe_ratio, stats


class OutPut(OnePyEnvBase):

    def show_setting(self, check_only=False):
        show_list = [(self.env.readers, 'readers'),
                     (self.env.cleaners, 'cleaners'),
                     (self.env.strategies, 'strategy'),
                     (self.env.brokers, 'brokers'),
                     (self.env.risk_managers, 'risk_managers'),
                     (self.env.recorders, 'recorders')]
        [check_setting(show, name, check_only) for show, name in show_list]

    def summary(self):
        """在最后输出简要回测结果"""
        total = pd.DataFrame(list(self.env.recorder.balance))
        total.set_index("date", inplace=True)
        pct_returns = total.pct_change()
        total = total / self.env.recorder.initial_cash
        max_drawdown, duration = create_drawdowns(total["value"])
        initial_cash = self.env.recorder.initial_cash
        final_value = self.env.recorder.balance.latest()
        sharpe_ratio = create_sharpe_ratio(pct_returns, len(pct_returns))
        d = {}
        d["Final_Value"] = f'${final_value:.2f}'
        d["Total_return"] = f'{(final_value/initial_cash-1)*100:.5f}%'
        d["Max_Drawdown"], d["Duration"] = f'{max_drawdown*100:.5f}%', duration
        d["Sharpe_Ratio"] = f'{sharpe_ratio:.5f}'
        print(dict_to_table(d))

    def plot(self, ticker):
        plotter = Plotly()

        return plotter.plot(ticker)

    def trade_log(self):
        return self.env.recorder.match_engine.generate_trade_log()




import queue

from OnePy.constants import EVENT


class Event(object):
    """TODO: 添加字典减少内存占用"""

    def __init__(self, event_type, **kwargs):
        self.__dict__ = kwargs
        self.event_type = event_type

    def __repr__(self):
        return ' '.join('{}:{}'.format(k, v) for k, v in self.__dict__.items())


class EventBus(object):
    def __init__(self):
        self.core = queue.Queue()

    def put(self, event):
        self.core.put(event)

    def get(self):
        return self.core.get(block=False)




from itertools import count

from OnePy.utils.awesome_func import dict_to_table


def check_setting(dict_data, name, check_only=False):
    if dict_data == {}:
        print('>'*10, f'Attention!! There is No {name}!!!', '<'*10)
    else:
        if not check_only:
            counter = count(1)
            print(dict_to_table({f'{name}_{next(counter)}': key
                                 for key in dict_data}))



from enum import Enum


class ActionType(Enum):

    Buy = 'Buy'
    Sell = 'Sell'
    Short_sell = 'Short_sell'
    Short_cover = 'Short_cover'

    Exit_all = 'Exit_all'
    Cancel_all = 'Cancel_all'


class OrderType(Enum):

    Market = 'Market'
    Limit = 'Limit'
    Stop = 'Stop'
    Trailing_stop = 'Trailing_stop'

    Limit_pct = 'Limit_pct'
    Stop_pct = 'Stop_pct'
    Trailing_stop_pct = 'Trailing_stop_pct'


class ExecType(Enum):

    Market_order = "MarketOrder"

    Limit_sell = 'Limit_sell'
    Limit_buy = 'Limit_buy'
    Stop_buy = 'Stop_buy'
    Stop_sell = 'Stop_sell'
    Trailing_stop_buy = 'Trailing_stop_buy'
    Trailing_stop_sell = 'Trailing_stop_sell'

    Limit_short_sell = 'Limit_short_sell'
    Stop_short_sell = 'Stop_short_sell'
    Limit_cover_short = 'Limit_cover_short'
    Stop_cover_short = 'Stop_cover_short'
    Trailing_stop_short_sell = 'Trailing_stop_short_sell'

    Exit_all = 'Exit_all'
    Close_all = "Close_all"


class OrderStatus(Enum):

    Created = "Created"
    Submitted = "Submitted"
    Partial = "Partial"
    Completed = "Completed"
    Canceled = "Canceled"
    Expired = "Expired"
    Margin = "Margin"
    Rejected = "Rejected"
    Triggered = "Triggered"


class EVENT(Enum):
    Market_updated = 'Market_updated'
    Data_cleaned = 'Data_cleaned'
    Signal_generated = 'Signal_generated'
    Submit_order = 'Submit_order'
    Record_result = 'Record_result'




import logging


class LoggerFactory:

    """
    CRITICAL	50
    ERROR	40
    WARNING	30
    INFO	20
    DEBUG	10
    NOTSET	0
    """

    def __init__(self, logger_name):
        self.logger = logging.getLogger(logger_name)
        self.logger_name = logger_name

        self.settle_info()
        self.settle_warning()

    def _set_file(self, logger_name, level_name, level, log_format):
        file_handler = logging.FileHandler(f'{logger_name}_{level_name}.log')

        formatter = logging.Formatter(log_format)
        file_handler.setLevel(level)
        file_handler.setFormatter(formatter)
        self.logger.addHandler(file_handler)

    def settle_info(self):
        self._set_file(self.logger_name, "INFO", logging.INFO, '%(message)s')

    def settle_warning(self):
        self._set_file(self.logger_name, "WARNING", logging.WARNING,
                       '%(asctime)s - %(filename)s[line:%(lineno)d] - %(levelname)s: %(message)s')



from OnePy.builtin_module.recorders.stock_recorder_series import (AvgPriceSeries,
                                                                  CommissionSeries,
                                                                  HoldingPnlSeries,
                                                                  MarginSeries,
                                                                  MarketValueSeries,
                                                                  PositionSeries,
                                                                  RealizedPnlSeries)
from OnePy.builtin_module.trade_log.stock_log import StockTradeLog
from OnePy.constants import ActionType
from OnePy.sys_module.base_recorder import RecorderBase
from OnePy.sys_module.models.base_series import CashSeries


class StockRecorder(RecorderBase):

    def __init__(self):
        super().__init__(StockTradeLog)

    def _for_long_or_short(self, order):
        if order.action_type in [ActionType.Buy, ActionType.Sell]:
            return 'long'
        elif order.action_type in [ActionType.Short_sell, ActionType.Short_cover]:
            return 'short'

    def _record_order(self):
        for order in self.env.orders_mkt_submitted:
            ticker = order.ticker
            long_or_short = self._for_long_or_short(order)

            last_position = self.position.latest(ticker, long_or_short)
            last_avg_price = self.avg_price.latest(ticker, long_or_short)
            last_commission = self.commission.latest(ticker, long_or_short)
            self.position.append(order, last_position, long_or_short)
            new_position = self.position.latest(ticker, long_or_short)

            self.avg_price.append(order,
                                  last_position,
                                  last_avg_price,
                                  new_position,
                                  long_or_short)
            new_avg_price = self.avg_price.latest(ticker, long_or_short)

            self.commission.append(order,
                                   last_commission,
                                   self.per_comm,
                                   self.per_comm_pct,
                                   long_or_short)
            self.realized_pnl.append(order,
                                     last_avg_price,
                                     new_avg_price,
                                     long_or_short)
            self.match_engine.match_order(order)
            self.update(order_executed=True)

    def _update_balance_and_cash(self, trading_date):
        total_realized_pnl = self.realized_pnl.total_value()
        total_holding_pnl = self.holding_pnl.total_value()
        total_commission = self.commission.total_value()
        total_margin = self.margin.total_value()
        total_market_value = self.market_value.total_value()
        new_balance = self.initial_cash + total_realized_pnl + \
            total_holding_pnl - total_commission
        new_frozen_cash = total_margin+total_market_value  # 更新frozen_cash

        new_cash = self.initial_cash + total_realized_pnl - \
            total_commission - new_frozen_cash  # 更新cash

        self.balance.append(dict(date=trading_date, value=new_balance))
        self.frozen_cash.append(dict(date=trading_date, value=new_frozen_cash))
        self.cash.append(dict(date=trading_date, value=new_cash))

    def set_setting(self, initial_cash=100000,
                    comm=1, comm_pct=None, margin_rate=0.1):
        self.initial_cash = initial_cash
        self.per_comm = comm
        self.per_comm_pct = comm_pct
        self.margin_rate = margin_rate

    def initialize(self):
        self.position = PositionSeries()
        self.avg_price = AvgPriceSeries()
        self.holding_pnl = HoldingPnlSeries()
        self.realized_pnl = RealizedPnlSeries()
        self.commission = CommissionSeries()
        self.market_value = MarketValueSeries()
        self.margin = MarginSeries()

        self.cash = CashSeries('cash', self.initial_cash)
        self.frozen_cash = CashSeries('frozen_cash', 0)
        self.balance = CashSeries('balance', self.initial_cash)

    def update(self, order_executed=False):
        """根据最新价格更新信息,
        需要更新cash，frozen cash, market_value, holding_pnl, balance"""
        self.market_value.update_barly(order_executed)
        self.holding_pnl.update_barly(order_executed)
        self.margin.update_barly(order_executed)
        self._update_balance_and_cash(self.env.trading_datetime)

    def run(self):
        self._record_order()



import multiprocessing
import multiprocessing.dummy as threading
import sys
import time
from functools import wraps

from retry import retry
from six import iteritems


def run_multiprocessing(func, params: list, max_worker: int, join=True):
    pool = multiprocessing.Pool(max_worker)

    if isinstance(params[0], tuple):
        pool.starmap(func, params)
    else:
        pool.map(func, params)
    pool.close()

    if join:
        pool.join()


def run_multithreading(func, params: list, max_worker: int, join=True):
    pool = threading.Pool(max_worker)

    if isinstance(params[0], tuple):
        pool.starmap(func, params)
    else:
        pool.map(func, params)
    pool.close()

    if join:
        pool.join()


def show_process(finished_len, total_len):
    i = int(finished_len/total_len*100)
    k = i + 1
    output = '>'*(i//2)+' '*((100-k)//2)
    sys.stdout.write('\r'+output+'[%s%%]' % (i+1))
    sys.stdout.flush()


def dict_to_table(_dict):
    col_width_keys = max([len(key) for key in _dict.keys()])
    col_width_values = max([len(str(value)) for value in _dict.values()])
    table_end_str = ("+-" + (col_width_keys +
                             col_width_values + 3) * "-" + "-+")
    data_strs = []

    for key, value in iteritems(_dict):
        data_strs.append("| " + " | ".join([
            "{:{}}".format(key, col_width_keys),
            "{:{}}".format(value, col_width_values)]) + " |\n")

    return "%s\n%s%s" % (table_end_str, "".join(data_strs), table_end_str)


def run_fuction(*funcs):
    for func in funcs:
        print(20*'=', f'Function: {func.__name__}', 20*'=')
        begin = time.time()
        func()
        end = time.time()
        print(f'Finished in {end-begin:.2f}s')


def handle_error(tries=20, delay=0.01):

    def decorate(func):

        @wraps(func)
        @retry(tries=tries, delay=delay)
        def wrapper(*args, **kwargs):
            try:
                func(*args, **kwargs)
            except Exception as error:
                print(f'Raise an error: {error}')
                print(f'Start to retry!!!!!')
                raise Exception

        return wrapper

    return decorate


if __name__ == "__main__":

    # Functions below are just for test.
    def just_sleep(index, test_list):
        i = len(test_list)
        time.sleep(0.01)
        show_process(i, 200)
        test_list.append(index)

    def test_multithreading():
        print('Start test MultiThreading')
        finished_list = multiprocessing.Manager().list()
        run_multithreading(
            just_sleep, [(i, finished_list) for i in range(200)], 2)

    def test_multiprocessing():
        print('Start test MultiProcessing')
        finished_list = multiprocessing.Manager().list()
        run_multiprocessing(
            just_sleep, [(i, finished_list) for i in range(200)], 2)

    run_fuction(test_multithreading, test_multiprocessing)



# encoding: UTF-8

"""
OnePy - An Own Built Trading Framework in Python.

The OnePy project is an open-source quantitative trading framework.
The project is mainly written in Python.

"""


import os

from setuptools import Command, find_packages, setup


class CleanCommand(Command):
    """Custom clean command to tidy up the project root."""
    user_options = []

    def initialize_options(self):
        pass

    def finalize_options(self):
        pass

    def run(self):
        os.system("rm -vrf ./build ./dist ./*.pyc ./*.egg-info")


setup(
    name="Onepy_trader",
    version="2.1",
    author="Chandler_Chan",
    author_email="chenjiayicjy@gmail.com",
    license="MIT",
    url="https://github.com/Chandlercjy/OnePy",
    description="Python Backtesting library for OnePiece.",
    packages=find_packages(),
    cmdclass={
        "clean": CleanCommand,
    }, install_requires=["pandas", "numpy", "plotly", "funcy", "arrow", "pymongo"]
)



from functools import wraps
import arrow


def make_it_float(func):
    @wraps(func)
    def wrapper(*args, **kargs):
        return float(func(*args, **kargs))

    return wrapper


def make_it_datetime(func):
    @wraps(func)
    def wrapper(*args, **kargs):
        return arrow.get(func(*args, **kargs))

    return wrapper



from collections import defaultdict

from OnePy.constants import ActionType, OrderStatus
from OnePy.sys_module.components.signal_generator import \
    TriggeredSignalGenerator
from OnePy.sys_module.metabase_env import OnePyEnvBase


class PendingOrderChecker(OnePyEnvBase):

    def _check_orders_pending(self):
        for order in self.env.orders_pending[:]:
            if self._send_signal(order):
                order.status = OrderStatus.Triggered
                self.env.orders_pending.remove(order)  # 成交的单子需要删除

    def _check_orders_pending_with_mkt(self):
        for key in list(self.env.orders_pending_mkt_dict):
            for order in self.env.orders_pending_mkt_dict[key]:

                if self._send_signal(order):

                    # TODO: 有Flaw，当多个pending order同时触，只用了第一个
                    order.status = OrderStatus.Triggered
                    del self.env.orders_pending_mkt_dict[key]

                    break

    def _send_signal(self, order):
        if TriggeredSignalGenerator.generate_triggered_signal(order):
            return True

    def run(self):
        self._check_orders_pending_with_mkt()
        self._check_orders_pending()


class SubmitOrderChecker(OnePyEnvBase):
    """可能有股票停牌情况"""
    """重新分一下发送订单和检查信号之间的关系"""

    def __init__(self, required_cash_func):
        self.required_cash_func = required_cash_func
        self.cash_acumu = None
        self.plong_acumu = None
        self.pshort_acumu = None

    @property
    def cur_cash(self):
        return self.env.recorder.cash[-1]['value']

    def required_cash(self, order):
        return self.required_cash_func(order)

    def _lack_of_cash(self, order):  # 用于Buy和Short Sell
        return True if self.cash_acumu > self.cur_cash else False

    def _lack_of_position(self, cur_position, acumu_position):  # 用于Sell指令和Cover指令
        if cur_position == 0 or acumu_position > cur_position:
            return True

        return False

    def add_to_position_cumu(self, order):

        if order.action_type == ActionType.Sell:
            self.plong_acumu[order.ticker] += order.size
        elif order.action_type == ActionType.Short_cover:
            self.pshort_acumu[order.ticker] += order.size

    def add_to_cash_cumu(self, order):
        self.cash_acumu += self.required_cash(order)

    def delete_from_cash_cumu(self, order):
        self.cash_acumu -= self.required_cash(order)

    def delete_from_position_cumu(self, order):

        if order.action_type == ActionType.Sell:
            self.plong_acumu[order.ticker] -= order.size
        elif order.action_type == ActionType.Short_cover:
            self.pshort_acumu[order.ticker] -= order.size

    def make_position_cumu_full(self, order):
        if order.action_type == ActionType.Sell:
            self.plong_acumu[order.ticker] = self.cur_position(order)
        elif order.action_type == ActionType.Short_cover:
            self.pshort_acumu[order.ticker] = self.cur_position(order)

    def cur_position(self, order):
        """根据order自动判断需要选取long还是short的position"""

        if order.action_type == ActionType.Sell:
            return self.env.recorder.position.latest(order.ticker, 'long')
        elif order.action_type == ActionType.Short_cover:
            return self.env.recorder.position.latest(order.ticker, 'short')

    def acumu_position(self, order):
        if order.action_type == ActionType.Sell:
            return self.plong_acumu[order.ticker]
        elif order.action_type == ActionType.Short_cover:
            return self.pshort_acumu[order.ticker]

    def order_pass_checker(self, order):
        order.status = OrderStatus.Submitted
        self.env.orders_mkt_submitted.append(order)

    def is_partial(self, order, cur_position, acumu_position):
        diff = cur_position-(acumu_position-order.size)

        if diff > 0:
            order.size = diff
            order.status = OrderStatus.Partial

            return True

    def _check(self, order_list):

        for order in order_list:

            if order.action_type in [ActionType.Buy, ActionType.Short_sell]:
                self.add_to_cash_cumu(order)

                if self._lack_of_cash(order):
                    order.status = OrderStatus.Rejected
                    self.delete_from_cash_cumu(order)

                    if order.mkt_id in self.env.orders_pending_mkt_dict:
                        del self.env.orders_pending_mkt_dict[order.mkt_id]

                    continue
            elif order.action_type in [ActionType.Sell, ActionType.Short_cover]:
                self.add_to_position_cumu(order)

                cur_position = self.cur_position(order)
                acumu_position = self.acumu_position(order)

                if self._lack_of_position(cur_position, acumu_position):
                    if self.is_partial(order, cur_position, acumu_position):
                        self.make_position_cumu_full(order)
                    else:
                        order.status = OrderStatus.Rejected
                        self.delete_from_position_cumu(order)

                        continue

            self.order_pass_checker(order)

    def _check_market_order(self):
        self.cash_acumu = 0
        self.plong_acumu = defaultdict(int)
        self.pshort_acumu = defaultdict(int)

        self._check(self.env.orders_mkt_absolute)
        self._check(self.env.orders_mkt_normal)

    def _check_pending_order(self):
        pass  # TODO: 实盘需要检查

    def _clear_all_mkt_order(self):
        self.env.orders_mkt_absolute = []
        self.env.orders_mkt_normal = []

    def run(self):
        self._check_market_order()
        self._check_pending_order()
        self._clear_all_mkt_order()



import logging
import queue

from OnePy.config import EVENT_LOOP
from OnePy.environment import Environment
from OnePy.event import Event
from OnePy.sys_module.components.exceptions import BacktestFinished
from OnePy.sys_module.components.logger import LoggerFactory
from OnePy.sys_module.components.market_maker import MarketMaker
from OnePy.sys_module.components.order_checker import PendingOrderChecker
from OnePy.sys_module.components.output import OutPut
from OnePy.sys_module.metabase_env import OnePyEnvBase


class OnePiece(object):

    env = Environment()

    def __init__(self):
        self.market_maker = MarketMaker()
        self.order_checker = PendingOrderChecker()
        self.cur_event = None
        self.env.logger = logging.getLogger("OnePy")

    def sunny(self, summary=True):
        """主循环，OnePy的核心"""
        self._initialize_trading_system()

        while True:
            try:
                self.cur_event = self.env.event_bus.get()
            except queue.Empty:
                try:
                    self.market_maker.update_market()
                    self.order_checker.run()
                except BacktestFinished:
                    self.output.summary() if summary else None

                    break
            else:
                self._run_event_loop()

    def _run_event_loop(self):
        for element in self.env.event_loop:
            if self._event_is_executed(**element):
                break

    def _event_is_executed(self, if_event, then_event, module_dict):
        if self.cur_event.event_type == if_event:
            [value.run() for value in module_dict.values()]
            self.env.event_bus.put(Event(then_event)) if then_event else None

            return True

    def _initialize_trading_system(self):
        self.env.refresh()
        OnePyEnvBase.env = self.env
        self.env.event_loop = EVENT_LOOP
        self.market_maker.initialize()

        if self.env.recorder:
            self.env.recorder.initialize()

    @property
    def output(self):
        return OutPut()

    def show_log(self, file=False):
        if file:
            LoggerFactory("OnePy").logger
        logging.basicConfig(level=logging.INFO)



import math
import operator
from collections import OrderedDict
from datetime import datetime

import numpy as np
import pandas as pd
from dateutil.relativedelta import relativedelta
from numpy.lib.stride_tricks import as_strided

#####################################################################
# CONSTANTS

TRADING_DAYS_PER_YEAR = 252
TRADING_DAYS_PER_MONTH = 20
TRADING_DAYS_PER_WEEK = 5


#####################################################################
# HELPER FUNCTIONS

def create_sharpe_ratio(returns, periods=252):
    """
    Create the Sharpe ratio for the strategy, based on a
    benchmark of zero (i.e. no risk-free rate information).

    Parameters:

    returns - A pandas Series representing period percentage returns.
    periods - Daily (252), Hourly (252*6.5), Minutely(252*6.5*60) etc.
    """
    ratio = np.sqrt(periods) * (np.mean(returns)) / np.std(returns)

    return ratio[0]


def create_drawdowns(equity_curve):
    """
    Calculate the largest peak-to-trough drawdown of the PnL curve
    as well as the duration of the drawdown. Requires that the
    pnl_returns is a pandas Series.

    Parameters:
    pnl - A pandas Series representing period percentage returns.

    Returns:
    drawdown, duration - Highest peak-to-trough drawdown and duration.
    """

    # Calculate the cumulative returns curve
    # and set up the High Water Mark
    # Then create the drawdown and duration series
    equity_curve.reset_index(drop=True, inplace=True)
    hwm = [0]
    eq_idx = equity_curve.index
    drawdown = pd.Series(index=eq_idx)
    duration = pd.Series(index=eq_idx)

    # Loop over the index range

    for t in range(1, len(eq_idx)):
        cur_hwm = max(hwm[t - 1], equity_curve[t])
        hwm.append(cur_hwm)
        drawdown[t] = hwm[t] - equity_curve[t]
        duration[t] = 0 if drawdown[t] == 0 else duration[t - 1] + 1

    return round(drawdown.max(), 5), round(duration.max(), 3)


def _difference_in_years(start, end):
    """ calculate the number of years between two dates """
    diff = end - start
    diff_in_years = (diff.days + diff.seconds / 86400) / 365.2425

    return diff_in_years


#####################################################################
# OVERALL RESULTS

def beginning_balance(capital):
    return capital


def ending_balance(dbal):
    return dbal.iloc[-1]['balance']


def unrealized_profit(tlog, dbal, capital):
    return ending_balance(dbal) - tlog.iloc[-1]['cumul_total'] - beginning_balance(capital)


def total_net_profit(tlog, dbal, capital):
    return tlog.iloc[-1]['cumul_total'] + unrealized_profit(tlog, dbal, capital)


def gross_profit(tlog, dbal, capital):
    un = unrealized_profit(tlog, dbal, capital)

    if un >= 0:
        return tlog[tlog['re_profit'] > 0].sum()['re_profit'] + un
    else:
        return tlog[tlog['re_profit'] > 0].sum()['re_profit']


def gross_loss(tlog, dbal, capital):
    un = unrealized_profit(tlog, dbal, capital)

    if un <= 0:
        return tlog[tlog['re_profit'] < 0].sum()['re_profit'] + un
    else:
        return tlog[tlog['re_profit'] < 0].sum()['re_profit']


def profit_factor(tlog, dbal, capital):
    if gross_profit(tlog, dbal, capital) == 0:
        return 0

    if gross_loss(tlog, dbal, capital) == 0:
        return 1000

    return gross_profit(tlog, dbal, capital) / gross_loss(tlog, dbal, capital) * -1


def return_on_initial_capital(tlog, dbal, capital):
    return total_net_profit(tlog, dbal, capital) / capital * 100


def _cagr(B, A, n):
    """ calculate compound annual growth rate """

    return (math.pow(B / A, 1 / n) - 1) * 100


def annual_return_rate(end_balance, capital, start, end):
    B = end_balance
    A = capital
    n = _difference_in_years(start, end)

    return _cagr(B, A, n)


def trading_period(start, end):
    diff = relativedelta(end, start)

    return '{} years {} months {} days'.format(diff.years, diff.months,
                                               diff.days)


def _true_func(arg1, arg2):
    return True


def _total_days_in_market(ts, tlog):
    l = _get_trade_bars(ts, tlog, _true_func)

    return sum(l)


def pct_time_in_market(ts, tlog, start, end):
    return _total_days_in_market(ts, tlog) / len(ts[start:end].index) * 100


#####################################################################
# SUMS

def total_num_trades(tlog):
    return len(tlog.index)


def num_winning_trades(tlog):
    return (tlog['re_profit'] > 0).sum()


def num_losing_trades(tlog):
    return (tlog['re_profit'] < 0).sum()


def num_even_trades(tlog):
    return (tlog['re_profit'] == 0).sum()


def pct_profitable_trades(tlog):
    if total_num_trades(tlog) == 0:
        return 0

    return num_winning_trades(tlog) / total_num_trades(tlog) * 100


#####################################################################
# CASH PROFITS AND LOSSES

def avg_profit_per_trade(tlog, dbal, capital):
    if total_num_trades(tlog) == 0:
        return 0

    return total_net_profit(tlog, dbal, capital) / total_num_trades(tlog)


def avg_profit_per_winning_trade(tlog, dbal, capital):
    if num_winning_trades(tlog) == 0:
        return 0

    return gross_profit(tlog, dbal, capital) / num_winning_trades(tlog)


def avg_loss_per_losing_trade(tlog, dbal, capital):
    if num_losing_trades(tlog) == 0:
        return 0

    return gross_loss(tlog, dbal, capital) / num_losing_trades(tlog)


def ratio_avg_profit_win_loss(tlog, dbal, capital):
    if avg_profit_per_winning_trade(tlog, dbal, capital) == 0:
        return 0

    if avg_loss_per_losing_trade(tlog, dbal, capital) == 0:
        return 1000

    return (avg_profit_per_winning_trade(tlog, dbal, capital) /
            avg_loss_per_losing_trade(tlog, dbal, capital) * -1)


def largest_profit_winning_trade(tlog):
    if num_winning_trades(tlog) == 0:
        return 0

    return tlog[tlog['re_profit'] > 0].max()['re_profit']


def largest_loss_losing_trade(tlog):
    if num_losing_trades(tlog) == 0:
        return 0

    return tlog[tlog['re_profit'] < 0].min()['re_profit']


#####################################################################
# POINTS

def num_winning_points(tlog):
    if num_winning_trades(tlog) == 0:
        return 0

    return tlog[tlog['pl_points'] > 0].sum()['pl_points']


def num_losing_points(tlog):
    if num_losing_trades(tlog) == 0:
        return 0

    return tlog[tlog['pl_points'] < 0].sum()['pl_points']


def total_net_points(tlog):
    return num_winning_points(tlog) + num_losing_points(tlog)


def avg_points(tlog):
    if total_num_trades(tlog) == 0:
        return 0

    return tlog['pl_points'].sum() / len(tlog.index)


def largest_points_winning_trade(tlog):
    if num_winning_trades(tlog) == 0:
        return 0

    return tlog[tlog['pl_points'] > 0].max()['pl_points']


def largest_points_losing_trade(tlog):
    if num_losing_trades(tlog) == 0:
        return 0

    return tlog[tlog['pl_points'] < 0].min()['pl_points']


def avg_pct_gain_per_trade(tlog):
    if total_num_trades(tlog) == 0:
        return 0
    df = tlog['pl_points'] / tlog['entry_price']

    return np.average(df) * 100


def largest_pct_winning_trade(tlog):
    if num_winning_trades(tlog) == 0:
        return 0
    df = tlog[tlog['pl_points'] > 0]
    df = df['pl_points'] / df['entry_price']

    return df.max() * 100


def largest_pct_losing_trade(tlog):
    if num_losing_trades(tlog) == 0:
        return 0
    df = tlog[tlog['pl_points'] < 0]
    df = df['pl_points'] / df['entry_price']

    return df.min() * 100


#####################################################################
# STREAKS

def _subsequence(s, c):
    """
    Takes as parameter list like object s and returns the length of the longest
    subsequence of s constituted only by consecutive character 'c's.
    Example: If the string passed as parameter is "001000111100", and c is '0',
    then the longest subsequence of only '0's has length 3.
    """

    bit = 0  # current element in the sequence
    count = 0  # current length of the sequence of zeros
    maxlen = 0  # temporary value of the maximum length

    for i in range(len(s)):
        bit = s[i]

        if bit == c:  # we have read a new '0'
            count = count + 1  # update the length of the current sequence

            if count > maxlen:  # if necessary, ...
                # ... update the temporary maximum
                maxlen = count
        else:  # we have read a 1
            count = 0  # reset the length of the current sequence

    return maxlen


def max_consecutive_winning_trades(tlog):
    if num_winning_trades(tlog) == 0:
        return 0

    return _subsequence(tlog['re_profit'] > 0, True)


def max_consecutive_losing_trades(tlog):
    if num_losing_trades(tlog) == 0:
        return 0

    return _subsequence(tlog['re_profit'] > 0, False)


def avg_bars_winning_trades(ts, tlog):
    if num_winning_trades(tlog) == 0:
        return 0

    return np.average(_get_trade_bars(ts, tlog, operator.gt))


def avg_bars_losing_trades(ts, tlog):
    if num_losing_trades(tlog) == 0:
        return 0

    return np.average(_get_trade_bars(ts, tlog, operator.lt))


#####################################################################
# DRAWDOWN AND RUNUP

def max_closed_out_drawdown(close):
    """ only compare each point to the previous running peak O(N) """
    running_max = close.expanding().max()
    cur_dd = (close - running_max) / running_max * 100
    dd_max = min(0, cur_dd.min())
    idx = cur_dd.idxmin()

    dd = pd.Series()
    dd['max'] = dd_max
    dd['peak'] = running_max[idx]
    dd['trough'] = close[idx]

    dd['start_date'] = close[close == dd['peak']
                             ].index[0].strftime("%Y-%m-%d %H:%M:%S")
    dd['end_date'] = idx.strftime("%Y-%m-%d %H:%M:%S")
    close = close[close.index > idx]

    rd_mask = close > dd['peak']

    if rd_mask.any():
        dd['recovery_date'] = \
            close[rd_mask].index[0].strftime("%Y-%m-%d %H:%M:%S")
    else:
        dd['recovery_date'] = 'Not Recovered Yet'

    return dd


def max_intra_day_drawdown(high, low):
    """ only compare each point to the previous running peak O(N) """
    running_max = high.expanding().max()
    cur_dd = (low - running_max) / running_max * 100
    dd_max = min(0, cur_dd.min())
    idx = cur_dd.idxmin()

    dd = pd.Series()
    dd['max'] = dd_max
    dd['peak'] = running_max[idx]
    dd['trough'] = low[idx]
    dd['start_date'] = high[high == dd['peak']
                            ].index[0].strftime("%Y-%m-%d %H:%M:%S")
    dd['end_date'] = idx.strftime("%Y-%m-%d %H:%M:%S")
    high = high[high.index > idx]

    rd_mask = high > dd['peak']

    if rd_mask.any():
        dd['recovery_date'] = \
            high[rd_mask].index[0].strftime("%Y-%m-%d %H:%M:%S")

    return dd


def _windowed_view(x, window_units):
    """Create a 2d windowed view of a 1d array.

    `x` must be a 1d numpy array.

    `numpy.lib.stride_tricks.as_strided` is used to create the view.
    The data is not copied.

    Example:

    >>> x = np.array([1, 2, 3, 4, 5, 6])
    >>> _windowed_view(x, 3)
    array([[1, 2, 3],
           [2, 3, 4],
           [3, 4, 5],
           [4, 5, 6]])
    """
    y = as_strided(x, shape=(x.size - window_units + 1, window_units),
                   strides=(x.strides[0], x.strides[0]))

    return y


def rolling_max_dd(ser, period, min_periods=1):
    """Compute the rolling maximum drawdown of `ser`.

    `ser` must be a Series.
    `min_periods` should satisfy `1 <= min_periods <= window_units`.

    Returns an 1d array with length `len(x) - min_periods + 1`.
    """
    window_units = period + 1
    x = ser.values

    if min_periods < window_units:
        pad = np.empty(window_units - min_periods)
        pad.fill(x[0])
        x = np.concatenate((pad, x))
    y = _windowed_view(x, window_units)
    running_max_y = np.maximum.accumulate(y, axis=1)
    dd = (y - running_max_y) / running_max_y * 100
    rmdd = dd.min(axis=1)

    return pd.Series(data=rmdd, index=ser.index, name=ser.name)


def rolling_max_ru(ser, period, min_periods=1):
    """Compute the rolling maximum runup of `ser`.

    `ser` must be a Series.
    `min_periods` should satisfy `1 <= min_periods <= window_units`.

    Returns an 1d array with length `len(x) - min_periods + 1`.
    """
    window_units = period + 1
    x = ser.values

    if min_periods < window_units:
        pad = np.empty(window_units - min_periods)
        pad.fill(x[0])
        x = np.concatenate((pad, x))
    y = _windowed_view(x, window_units)
    running_min_y = np.minimum.accumulate(y, axis=1)
    ru = (y - running_min_y) / running_min_y * 100
    rmru = ru.max(axis=1)

    return pd.Series(data=rmru, index=ser.index, name=ser.name)


#####################################################################
# PERCENT CHANGE - used to compute several stastics

def pct_change(close, period):
    diff = (close.shift(-period) - close) / close * 100
    diff.dropna(inplace=True)

    return diff


#####################################################################
# RATIOS

def sharpe_ratio(rets, risk_free=0.00, period=TRADING_DAYS_PER_YEAR):
    """
    summary Returns the daily Sharpe ratio of the returns.
    param rets: 1d numpy array or fund list of daily returns (centered on 0)
    param risk_free: risk free returns, default is 0%

    return Sharpe Ratio, computed off daily returns
    """
    dev = np.std(rets, axis=0)
    mean = np.mean(rets, axis=0)
    sharpe = (mean * period - risk_free) / (dev * np.sqrt(period))

    return sharpe


def sortino_ratio(rets, risk_free=0.00, period=TRADING_DAYS_PER_YEAR):
    """
    summary Returns the daily Sortino ratio of the returns.
    param rets: 1d numpy array or fund list of daily returns (centered on 0)
    param risk_free: risk free return, default is 0%

    return Sortino Ratio, computed off daily returns
    """
    mean = np.mean(rets, axis=0)
    negative_rets = rets[rets < 0]
    dev = np.std(negative_rets, axis=0)
    sortino = (mean * period - risk_free) / (dev * np.sqrt(period))

    return sortino


#####################################################################
# STATS - this is the primary call used to generate the results

def stats(ts, tlog, dbal, start, end, capital):
    """
    Compute trading stats
    Parameters
    ----------
    ts : Dataframe
        Time series of security prices (date, high, low, close, volume,
        adj_close)
    tlog : Dataframe
        Trade log (entry_date, entry_price, long_short, qty,
        exit_date, exit_price, pl_points, re_profit, cumul_total)
    dbal : Dataframe
        Daily Balance (date, high, low, close)
    start : datetime
        date of first buy
    end : datetime
        date of last sell
    capital : float
        starting capital

    Examples
    --------

    Returns
    -------
    stats : Series of stats
    """

    stats = OrderedDict()

    # OVERALL RESULTS
    stats['start'] = start.strftime("%Y-%m-%d %H:%M:%S")
    stats['end'] = end.strftime("%Y-%m-%d %H:%M:%S")
    stats['beginning_balance'] = beginning_balance(capital)
    stats['ending_balance'] = ending_balance(dbal)
    stats['unrealized_profit'] = unrealized_profit(tlog, dbal, capital)
    stats['total_net_profit'] = total_net_profit(tlog, dbal, capital)
    stats['gross_profit'] = gross_profit(tlog, dbal, capital)
    stats['gross_loss'] = gross_loss(tlog, dbal, capital)
    stats['profit_factor'] = profit_factor(tlog, dbal, capital)
    stats['return_on_initial_capital'] = \
        return_on_initial_capital(tlog, dbal, capital)
    cagr = annual_return_rate(dbal['balance'][-1], capital, start, end)
    stats['annual_return_rate'] = cagr
    stats['trading_period'] = trading_period(start, end)
    stats['pct_time_in_market'] = \
        pct_time_in_market(ts, tlog, start, end)

    # SUMS
    stats['total_num_trades'] = total_num_trades(tlog)
    stats['num_winning_trades'] = num_winning_trades(tlog)
    stats['num_losing_trades'] = num_losing_trades(tlog)
    stats['num_even_trades'] = num_even_trades(tlog)
    stats['pct_profitable_trades'] = pct_profitable_trades(tlog)

    # CASH PROFITS AND LOSSES
    stats['avg_profit_per_trade'] = avg_profit_per_trade(tlog, dbal, capital)
    stats['avg_profit_per_winning_trade'] = avg_profit_per_winning_trade(
        tlog, dbal, capital)
    stats['avg_loss_per_losing_trade'] = avg_loss_per_losing_trade(
        tlog, dbal, capital)
    stats['ratio_avg_profit_win_loss'] = ratio_avg_profit_win_loss(
        tlog, dbal, capital)
    stats['largest_profit_winning_trade'] = largest_profit_winning_trade(tlog)
    stats['largest_loss_losing_trade'] = largest_loss_losing_trade(tlog)

    # POINTS
    stats['num_winning_points'] = num_winning_points(tlog)
    stats['num_losing_points'] = num_losing_points(tlog)
    stats['total_net_points'] = total_net_points(tlog)
    stats['avg_points'] = avg_points(tlog)
    stats['largest_points_winning_trade'] = largest_points_winning_trade(tlog)
    stats['largest_points_losing_trade'] = largest_points_losing_trade(tlog)
    stats['avg_pct_gain_per_trade'] = avg_pct_gain_per_trade(tlog)
    stats['largest_pct_winning_trade'] = largest_pct_winning_trade(tlog)
    stats['largest_pct_losing_trade'] = largest_pct_losing_trade(tlog)

    # STREAKS
    stats['max_consecutive_winning_trades'] = \
        max_consecutive_winning_trades(tlog)
    stats['max_consecutive_losing_trades'] = \
        max_consecutive_losing_trades(tlog)
    stats['avg_bars_winning_trades'] = \
        avg_bars_winning_trades(ts, tlog)
    stats['avg_bars_losing_trades'] = avg_bars_losing_trades(ts, tlog)

    # DRAWDOWN
    dd = max_closed_out_drawdown(dbal['balance'])
    stats['max_closed_out_drawdown'] = dd['max']
    stats['max_closed_out_drawdown_start_date'] = dd['start_date']
    stats['max_closed_out_drawdown_end_date'] = dd['end_date']
    stats['max_closed_out_drawdown_recovery_date'] = dd['recovery_date']
    stats['drawdown_recovery'] = _difference_in_years(
        datetime.strptime(dd['start_date'], "%Y-%m-%d %H:%M:%S"),
        datetime.strptime(dd['end_date'], "%Y-%m-%d %H:%M:%S")) * -1
    stats['drawdown_annualized_return'] = dd['max'] / cagr
    dd = max_intra_day_drawdown(dbal['balance_high'], dbal['balance_low'])
    stats['max_intra_day_drawdown'] = dd['max']
    dd = rolling_max_dd(dbal['balance'], TRADING_DAYS_PER_YEAR)
    stats['avg_yearly_closed_out_drawdown'] = np.average(dd)
    stats['max_yearly_closed_out_drawdown'] = min(dd)
    dd = rolling_max_dd(dbal['balance'], TRADING_DAYS_PER_MONTH)
    stats['avg_monthly_closed_out_drawdown'] = np.average(dd)
    stats['max_monthly_closed_out_drawdown'] = min(dd)
    dd = rolling_max_dd(dbal['balance'], TRADING_DAYS_PER_WEEK)
    stats['avg_weekly_closed_out_drawdown'] = np.average(dd)
    stats['max_weekly_closed_out_drawdown'] = min(dd)

    # RUNUP
    ru = rolling_max_ru(dbal['balance'], TRADING_DAYS_PER_YEAR)
    stats['avg_yearly_closed_out_runup'] = np.average(ru)
    stats['max_yearly_closed_out_runup'] = ru.max()
    ru = rolling_max_ru(dbal['balance'], TRADING_DAYS_PER_MONTH)
    stats['avg_monthly_closed_out_runup'] = np.average(ru)
    stats['max_monthly_closed_out_runup'] = max(ru)
    ru = rolling_max_ru(dbal['balance'], TRADING_DAYS_PER_WEEK)
    stats['avg_weekly_closed_out_runup'] = np.average(ru)
    stats['max_weekly_closed_out_runup'] = max(ru)

    # PERCENT CHANGE
    pc = pct_change(dbal['balance'], TRADING_DAYS_PER_YEAR)
    stats['pct_profitable_years'] = (pc > 0).sum() / len(pc) * 100
    stats['best_year'] = pc.max()
    stats['worst_year'] = pc.min()
    stats['avg_year'] = np.average(pc)
    stats['annual_std'] = pc.std()
    pc = pct_change(dbal['balance'], TRADING_DAYS_PER_MONTH)
    stats['pct_profitable_months'] = (pc > 0).sum() / len(pc) * 100
    stats['best_month'] = pc.max()
    stats['worst_month'] = pc.min()
    stats['avg_month'] = np.average(pc)
    stats['monthly_std'] = pc.std()
    pc = pct_change(dbal['balance'], TRADING_DAYS_PER_WEEK)
    stats['pct_profitable_weeks'] = (pc > 0).sum() / len(pc) * 100
    stats['best_week'] = pc.max()
    stats['worst_week'] = pc.min()
    stats['avg_week'] = np.average(pc)
    stats['weekly_std'] = pc.std()

    # RATIOS
    stats['sharpe_ratio'] = sharpe_ratio(dbal['balance'].pct_change())
    stats['sortino_ratio'] = sortino_ratio(dbal['balance'].pct_change())

    for i, j in stats.items():
        if type(j) is not str:
            stats[i] = round(j, 3)

    return stats


#####################################################################
# SUMMARY - stats() must be called before calling any of these functions

def summary(stats, *metrics):
    """ Returns stats summary in a DataFrame.
        stats() must be called before calling this function """
    index = []
    columns = ['strategy']
    data = []
    # add metrics

    for metric in metrics:
        index.append(metric)
        data.append(stats[metric])

    df = pd.DataFrame(data, columns=columns, index=index)

    return df


def summary2(stats, benchmark_stats, *metrics):
    """ Returns stats with benchmark summary in a DataFrame.
        stats() must be called before calling this function """
    index = []
    columns = ['strategy', 'benchmark']
    data = []
    # add metrics

    for metric in metrics:
        index.append(metric)
        data.append((stats[metric], benchmark_stats[metric]))

    df = pd.DataFrame(data, columns=columns, index=index)

    return df


def summary3(stats, benchmark_stats, *extras):
    """ Returns stats with benchmark summary in a DataFrame.
        stats() must be called before calling this function
        *extras: extra metrics """
    index = ['annual_return_rate',
             'max_closed_out_drawdown',
             'drawdown_annualized_return',
             'pct_profitable_months',
             'best_month',
             'worst_month',
             'sharpe_ratio',
             'sortino_ratio']
    columns = ['strategy', 'benchmark']
    data = [(stats['annual_return_rate'],
             benchmark_stats['annual_return_rate']),
            (stats['max_closed_out_drawdown'],
             benchmark_stats['max_closed_out_drawdown']),
            (stats['drawdown_annualized_return'],
             benchmark_stats['drawdown_annualized_return']),
            (stats['pct_profitable_months'],
             benchmark_stats['pct_profitable_months']),
            (stats['best_month'],
             benchmark_stats['best_month']),
            (stats['worst_month'],
             benchmark_stats['worst_month']),
            (stats['sharpe_ratio'],
             benchmark_stats['sharpe_ratio']),
            (stats['sortino_ratio'],
             benchmark_stats['sortino_ratio'])]
    # add extra metrics

    for extra in extras:
        index.append(extra)
        data.append((stats[extra], benchmark_stats[extra]))

    df = pd.DataFrame(data, columns=columns, index=index)

    return df



from OnePy.sys_module.components.cash_checker import CashChecker
from OnePy.sys_module.components.order_checker import SubmitOrderChecker
from OnePy.sys_module.components.order_generator import OrderGenerator
from OnePy.sys_module.metabase_env import OnePyEnvBase


class BrokerBase(OnePyEnvBase):

    def __init__(self):
        self.env.brokers.update({self.__class__.__name__: self})
        self._checker = None
        self._order_generator = OrderGenerator()

    def _clear_submited_order(self):
        self.env.orders_mkt_submitted = []

    def submit_order(self):
        pass

    def generate_order(self):
        self._order_generator.run()

    def run(self):
        self._clear_submited_order()
        self.generate_order()
        self.check_order()
        self.submit_order()

    def check_order(self):
        self._checker.run()


class StockBroker(BrokerBase):

    def __init__(self):
        super().__init__()
        self._checker = SubmitOrderChecker(CashChecker.stock_checker)



from collections import UserDict, UserList

import pandas as pd

from OnePy.constants import ActionType
from OnePy.sys_module.metabase_env import OnePyEnvBase


class SeriesBase(OnePyEnvBase, UserDict):

    def __init__(self):
        super().__init__()
        name = None

        for ticker in self.env.feeds:
            self.data[f'{ticker}_long'] = [
                dict(date=self.env.fromdate, value=0)]
            self.data[f'{ticker}_short'] = [
                dict(date=self.env.fromdate, value=0)]

    def latest(self, ticker, long_or_short, index=-1):
        return self.data[f'{ticker}_{long_or_short}'][index]['value']

    def total_value(self):
        total = 0

        for data_list in self.data.values():
            per_dict = data_list[-1]
            total += per_dict['value']

        return total

    def direction(self, order):
        if order.action_type in [ActionType.Buy, ActionType.Short_sell]:
            return 1

        elif order.action_type in [ActionType.Sell, ActionType.Short_cover]:
            return -1

    def earn_short(self, long_or_short):
        return 1 if long_or_short == 'long' else -1

    def _append_value(self, ticker, value, long_or_short):
        self.data[f'{ticker}_{long_or_short}'].append(
            {'date': self.env.trading_datetime, 'value': value})

    def dataframe(self):
        dataframe_list = []

        for ticker in self.env.tickers:
            long_df = pd.DataFrame(self.data[f'{ticker}_long'])
            short_df = pd.DataFrame(self.data[f'{ticker}_short'])
            long_df.rename(columns=dict(
                value=f'{self.name}_{ticker}_long'), inplace=True)
            short_df.rename(columns=dict(
                value=f'{self.name}_{ticker}_short'), inplace=True)

            dataframe_list.append(long_df)
            dataframe_list.append(short_df)

        dataframe = long_df

        for df in dataframe_list:
            dataframe = dataframe.merge(df, how='outer')

        dataframe.set_index('date', inplace=True)
        dataframe.index = pd.to_datetime(dataframe.index)
        dataframe.fillna(method='ffill', inplace=True)
        dataframe = dataframe[~dataframe.index.duplicated(keep='last')]

        return dataframe

    def plot(self, ticker):
        long_df = pd.DataFrame(self.data[f'{ticker}_long'])
        short_df = pd.DataFrame(self.data[f'{ticker}_short'])
        long_df.rename(columns=dict(value=f'{self.name}_long'), inplace=True)
        short_df.rename(columns=dict(value=f'{self.name}_short'), inplace=True)

        total_df = long_df.merge(short_df, how='outer')
        total_df.fillna(method='ffill', inplace=True)
        total_df.set_index('date', inplace=True)
        total_df.plot()

    def get_barly_cur_price(self, ticker, order_executed):
        if order_executed:
            return self.env.feeds[ticker].execute_price
        else:
            return self.env.feeds[ticker].open


class CashSeries(UserList):

    def __init__(self, name, initial_value):
        super().__init__()
        self.name = name
        self.data = []

    def latest(self):
        return self.data[-1]['value']

    def dataframe(self):
        dataframe = pd.DataFrame(self.data)
        dataframe.rename(columns=dict(value=self.name), inplace=True)
        dataframe.set_index('date', inplace=True)
        dataframe.index = pd.to_datetime(dataframe.index)
        result = dataframe[~dataframe.index.duplicated(keep='last')]

        first = dataframe.ix[:1]
        result = pd.concat([first, result])
        result.sort_index(inplace=True)

        return result

    def plot(self):
        self.dataframe().plot()



from itertools import count

import numpy as np

from OnePy.constants import OrderStatus, OrderType
from OnePy.sys_module.metabase_env import OnePyEnvBase
from OnePy.sys_module.models.signals import SignalByTrigger


class OrderBase(OnePyEnvBase):

    counter = count(1)

    def __init__(self, signal, mkt_id):
        self.signal = signal
        self.ticker = signal.ticker
        self.size = signal.size

        self.order_id = next(self.counter)
        self.mkt_id = mkt_id

        self.first_cur_price = self._get_first_cur_price()  # 记录订单发生时刻的现价

        self.status = OrderStatus.Created

    @property
    def trading_date(self):
        return self.signal.datetime

    def _get_first_cur_price(self):
        if self.signal.is_absolute_signal():
            return self.signal.execute_price

        return self.env.feeds[self.ticker].execute_price

    @property
    def action_type(self):
        return self.signal.action_type


class PendingOrderBase(OrderBase):

    def __init__(self, signal, mkt_id, trigger_key):
        self.trigger_key = trigger_key
        super().__init__(signal, mkt_id)

    @property
    def action_type(self):
        return self._action_type

    @property
    def status(self):
        return self._status

    @status.setter
    def status(self, value):
        self._status = value

        if self.signal.execute_price:
            execute_price = self.signal.execute_price
        else:
            execute_price = self.signal.price
        self.env.logger.info(
            f'{self.signal.datetime}, '
            f'{self.signal.ticker}, '
            f'{self.order_type.value} '
            f'{self.action_type.value} '
            f'@ {self.target_price:.5f}, '
            f'{self.status.value}, '
            f'size: {self.size}')

    def target_below(self) -> bool:
        raise NotImplementedError

    @property
    def cur_high(self):
        return self.env.feeds[self.ticker].high

    @property
    def cur_low(self):
        return self.env.feeds[self.ticker].low

    @property
    def money(self):
        return self.signal.get(self.trigger_key)

    @property
    def pct(self):
        if 'pct' in self.trigger_key:
            return self.signal.get(self.trigger_key)

    @property
    def difference(self):
        if self.pct:
            return abs(self.pct*self.first_cur_price)

        return abs(self.money/self.size)

    @property
    def target_price(self):
        if self.trigger_key == 'price':
            return self.signal.price

        elif self.target_below:

            return self.below_price(self.difference)

        return self.above_price(self.difference)

    @property
    def is_triggered(self):
        if self.target_below:
            return self.cur_low_cross_target_price()

        return self.cur_high_cross_target_price()

    def cur_high_cross_target_price(self):
        return True if self.target_price < self.cur_high else False

    def cur_low_cross_target_price(self):
        return True if self.target_price > self.cur_low else False

    def is_with_mkt(self):
        return False if self.trigger_key == 'price' else True

    def below_price(self, diff):
        return self.first_cur_price - diff

    def above_price(self, diff):
        return self.first_cur_price + diff


class TrailingOrderBase(PendingOrderBase):

    def __init__(self, signal, mkt_id, trigger_key):
        super().__init__(signal, mkt_id, trigger_key)
        self.latest_target_price = self.initialize_latest_target_price()

    @property
    def status(self):
        return self._status

    @status.setter
    def status(self, value):
        self._status = value

        if self.signal.execute_price:
            execute_price = self.signal.execute_price
        else:
            execute_price = self.signal.price
        self.env.logger.info(
            f'{self.signal.datetime}, '
            f'{self.signal.ticker}, '
            f'{self.order_type.value} '
            f'{self.action_type.value} '
            f'{self.status.value}, '
            f'size: {self.size}')

    @property
    def cur_open(self):
        return self.env.feeds[self.ticker].open

    def initialize_latest_target_price(self):
        if self.target_below:
            return self.first_cur_price - self.difference

        return self.first_cur_price + self.difference

    def target_below(self) -> bool:
        return False

    @property
    def difference(self):
        if self.pct:
            return abs(self.pct*self.first_cur_price)

        return abs(self.money/self.size)

    def with_high(self, diff):
        return np.random.uniform(self.cur_low, self.cur_high - diff)

    def with_low(self, diff):
        return np.random.uniform(self.cur_low + diff, self.cur_high)

    @property
    def target_price(self):

        if self.target_below:
            new = self.with_high(self.difference)  # TODO:有可能在触及高点前就成交了
            self.latest_target_price = max(self.latest_target_price, new)
        else:
            new = self.with_low(self.difference)  # TODO:有可能在触及高点前就成交了
            self.latest_target_price = min(self.latest_target_price, new)

        return self.latest_target_price
